from typing import Union
from .constants import FILTERS


def filter_exists(filter: str) -> bool:
    """Checks if a filter name exists

    Args:
        filter (str): filter name

    Returns:
        bool: whether or not the filter exists
    """
    return filter.lower() in list(FILTERS.keys())


def is_filter_value_allowed(filter: str, value: str) -> bool:
    """Checks if the value supplied exists for a filter

    Args:
        filter (str): filter name
        value (str): value/parameter of the filter

    Returns:
        bool: whether or not the value is allowed for the filter
    """
    if filter_exists(filter.lower()):
        return value.lower() in FILTERS[filter.lower()]["values"]

    else:
        raise ValueError(f"filter ({filter}) does not exist")


def prefixer(filter: str, value: str) -> str:
    """Prefixes the finviz code with the value to form part of the query string

    Args:
        filter (str): filter name
        value (str): value/parameter of the filter

    Returns:
        Union[str, None]: _description_
    """
    if filter_exists(filter.lower()) and is_filter_value_allowed(filter.lower(), value):
        return FILTERS[filter.lower()]["short"] + value.lower()
    
    else:
        raise ValueError(f"wrong combination of filter ({filter}) and value {value}")



from typing import List


def construct_table_from_data(headers: List[str], table: List[str]) -> list:
    data = []

    if len(table) < 1 or len(headers) < 1:
        return []
    
    number_of_rows = len(table)//len(headers)

    for i in range(number_of_rows):
        ticker_sub = {}
        row = table[i*len(headers):][:len(headers)]
        for j, header in enumerate(headers):
            ticker_sub[header] = row[j]

        data.append(ticker_sub)

    return data


filters = {
    "exchange": {
        "short": "exch_",
        "values": ["amex", "nasd", "nyse"]
    },

    "market cap": {
        "short": "cap_",
        "values": [
            "mega", "large", "mid", "small", "micro", "nano",
            "largeover", "midover", "smallover", "microover"
            "largeunder", "midunder", "smallunder", "microunder"

        ]
    },

    "earnings date": {
        "short": "earningsdate_",
        "values": [
            "today", "todaybefore", "todayafter",
            "tomorrow", "tomorrowbefore", "tomorrowafter",
            "yesterday", "yesterdaybefore", "yesterdayafter",
            "nextdays5", "prevdays5", "thisweek",
            "nextweek", "prevweek", "thismonth"
        ]

    },

    "target price": {
        "short": "targetprice_",
        "values": [
            "a50", "a40", "a30", "a20", "a10", "a5", "above", "below",
            "b5", "b10", "b20", "b30", "b40", "b50"
        ]
    },

    "index": {
        "short": "idx_",
        "values": [
            "sp500", "dji"
        ]
    },

    "dividend yield": {
        "short": "fa_div_",
        "values": [
            "none", "pos", "high", "veryhigh",
            "o1", "o2", "o3", "o4", "o5",
            "o6", "o7", "o8", "o9", "o10"
        ]
    },

    "average volume": {
        "short": "sh_avgvol_u500_",
        "values": [
            "u50", "u100", "u500", "u750", "u1000",
            "o50", "o100", "o500", "o750", "o1000", "o2000",
            "100to500", "100to1000", "500to1000", "500to10000"
        ]
    },

    "ipo date": {
        "short": "ipodate_",
        "values": [
            "today", "yesterday", "prevweek", "prevmonth", "prevquarter",
            "prevyear", "prev2years", "prev3years", "prev5years",
            "more1", "more5", "more10", "more15", "more20", "more25"
        ]
    },

    "sector": {
        "short": "sec_",
        "values": [
            "basicmaterials", "communicationservices", "consumercyclical", "consumerdefensive",
            "energy", "financial", "healthcare", "industrials", "realestate", "technology",
            "utilities",
        ]
    },

    "float short": {
        "short": "sh_short_",
        "values": [
            "low", "high",
            "u5", "u10", "u15", "u20", "u25", "u30",
            "o5", "o10", "o15", "020", "025", "030",
        ]
    },

    "relative volume": {
        "short": "sh_relvol_",
        "values": [
            "o10", "o5", "o3", "o3", "o2", "o1.5", "o1", "o0.75", "o0.5", "o0.25",
            "u2", "u1.5", "u1", "u0.75", "u0.5", "u0.25", "u0.1"
        ]
    },

    "shares outstanding": {
        "short": "sh_outstanding",
        "values": [
            "u1", "u5", "u10", "u20", "u50", "u100",
            "o1", "o2", "o5", "o10", "o20", "o50", "o100", "o200", "o500", "o1000"
        ]
    },

    "industry": {
        "short": "ind_",
        "values": [
            "stocksonly", "exchangetradedfund", "advertisingagencies", "aerospacedefense",
            "agriculturalinputs", "airlines", "airportsairservices", "aluminum", "apparelmanufacturing",
            "apparelretail", "assetmanagement", "automanufacturers", "autoparts", "autotruckdealerships",
            "banksdiversified", "banksregional", "beveragesbrewers", "beveragesnonalcoholic", "beverageswineriesdistilleries",
            "biotechnology", "broadcasting", "buildingmaterials", "buildingproductsequipment", "businessequipmentsupplies",
            "capitalmarkets", "chemicals", "closedendfunddebt", "closedendfundequity", "closedendfundforeign", "cokingcoal",
            "communicationequipment", "computerhardware", "confectioners", "conglomerates", "consultingservices", "consumerelectronics",
            "copper", "creditservices", "departmentstores", "diagnosticsresearch", "discountstores", "drugmanufacturersgeneral",
            "drugmanufacturersspecialtygeneric", "educationtrainingservices", "electricalequipmentparts", "electroniccomponents",
            "electronicgamingmultimedia", "electronicscomputerdistribution", "engineeringconstruction", "entertainment",
            "exchangetradedfund", "farmheavyconstructionmachinery", "farmproducts", "financialconglomerates",
            "financialdatastockexchanges", "fooddistribution", "footwearaccessories", "furnishingsfixturesappliances",
            "gambling", "gold", "grocerystores", "healthcareplans", "healthinformationservices", "homeimprovementretail",
            "householdpersonalproducts", "industrialdistribution"
        ]
    },

    "analyst recommendation": {
        "short": "an_recom_",
        "values": [
            "buybetter", "buy",
            "holdbetter", "hold", "holdworse",
            "sell", "sellworse", "strongsell",
        ]},

    "current volume": {
        "short": "sh_curvol_",
        "values": [
            "u50", "u100", "u500", "u750", "u1000",
            "o0", "o50", "o100", "o200", "o300", "o400", "o500", "o750",
            "o1000", "o2000", "o5000", "o10000", "o20000"
        ],
    },

    "float": {
        "short": "sh_float_",
        "values": [
            "u1", "u5", "u10", "u20", "u50", "u100",
            "o1", "o2", "o5", "o10", "o20", "o50", "o100", "o200", "o500", "o1000"
        ]
    },

    "country": {
        "short": "geo_",
        "values": [
            "usa", "not_usa", "asia", "europe", "latinamerica", "bric"
        ]
    },

    "option/short": {
        "short": "sh_opt_",
        "values": ["option", "short", "optionshort"],
    },

    "price": {
        "short": "sh_price",
        "values": [
            "u1", "u2", "u3", "u4", "u5", "u7", "u10", "u15", "u20", "u30", "u40", "u50",
            "o1", "o2", "o3", "o4", "o5", "o7", "o10", "o15", "o20", "o30", "o40", "o50",
            "o60", "o70", "o80", "o90", "o100",
            "1to5", "1to10", "1to20",
            "5to10", "5to20", "5to50", "10to20", "10to50", "20to50", "50to100"
        ]
    },

}



from typing import Optional
import requests
import random
from bs4 import BeautifulSoup
from .tickers import Tickers
from .filter import prefixer, is_filter_value_allowed, filter_exists
from .utils import get_number_of_pages, get_tickers_table, construct_table_from_data


class Screener:
    def __init__(self) -> None:
        """Screener that scrapes data from https://finviz.com
        """

        self.url = "https://finviz.com/screener.ashx?v=111"
        self.headers = [
            {
                'User-Agent': "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/W.X.Y.Z Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
            },
            {
                "User-Agent": "Googlebot/2.1 (+http://www.google.com/bot.html)",

            },

            {
                "User-Agent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/41.0.2272.96 Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",

            },

            {
                "User-Agent": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/107.0.5304.110 Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",

            }

        ]

    def get_header(self):
        return random.choice(self.headers)

    def query(self, filters: dict, limit: Optional[bool] = True) -> dict:

        url = self.url + self._construct_url(filters)
        data, total = self._get_page(url=url)

        tickers = Tickers(tickers=data)

        if limit and len(data) < limit:

            for page_idx in range(1, int(total)):
                tickers.add_tickers(self._get_page(url=url, page_idx=page_idx))
                if len(tickers) >= limit:
                    break

        return tickers

    def _get_page(self, url: str, page_idx: Optional[int] = None):
        """scrape data from a page given the page index and url

        Args:
            url (str): url of the page
            page_idx (Optional[int], optional): page index. Defaults to None.


        Returns:
            dict: data scraped from page
        """

        if page_idx and page_idx > 1:
            url += f"&r={20*(page_idx-1) + 1}"

        try:
            data = requests.get(
                url, headers=self.get_header()).content.decode()
            soup = BeautifulSoup(data, "html.parser")

            number_of_pages = get_number_of_pages(soup)

            # scrape data

            headers, table = get_tickers_table(soup)

            data = construct_table_from_data(headers, table)

            return data, number_of_pages

        except Exception as e:
            print("An error occured in geting the page...")
            raise e

    def _construct_url(self, filters: dict) -> str:
        """construct query url based on filters

        Args:
            filters (dict): dict of filters and the needed values

        Returns:
            str: query string
        """
        url = "&f="

        for category in filters:
            category_filters = filters[category]

            for filter_name in list(category_filters.keys()):

                if not filter_exists(filter_name):
                    raise ValueError(f"Invalid filter {filter_name}")

                filter_value = category_filters[filter_name]

                if not is_filter_value_allowed(filter_name, filter_value):
                    raise ValueError(
                        f"Invalid value `{filter_value}` for filter `{filter_name}`")

                url += prefixer(filter_name, filter_value) + ','

        url = url[:-1]  # remove that unecessary last ","

        return url



import setuptools

long_description = open("README.md", "r").read()
requirements = [i.strip() for i in open("requirements.txt", "r").readlines()]

setuptools.setup(
    name="fvscraper",
    version="0.0.1",
    author="Daniel Ik",
    description="Library for interacting with https://finviz.com",
    long_description_content_type='text/markdown',
    long_description=long_description,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    py_modules=["finscraper"],
    package_dir={"":"."},
    install_requires=requirements
)


from typing import Any, List, Optional


class Ticker:
    def __init__(self, data: dict) -> None:
        columns = list(data.keys())

        self.data = {}

        for column in columns:
            # make the column name friendlier
            new_key = column.lower()\
                .replace(" ", "_")\
                .replace(".", "")

            value = self._parse_value(data[column])

            self.data[new_key] = value

    def _parse_value(self, value: str) -> Any:
        if value == "-":
            return None

        if value.count(",") > 0:
            if value.replace(",", "").isnumeric():
                return int(value.replace(",", ""))
            else:
                return value

        if value.count(".") > 0:
            try:
                return float(value)
            except ValueError:
                try:
                    return float(
                        value
                        .replace("M", "E6")
                        .replace("B", "E9")
                        .replace("T", "E12")
                        .replace("%", "")
                    )
                except:
                    return value

        if value.count("%") > 0:
            return float(value.replace("%", ""))

        return value

    def _item(self):
        return self.data

    def _headers(self):
        return list(self.data.keys())

    def _values(self):
        return list(self.data.values())

    def __getattr__(self, __name: str) -> Any:
        try:
            value = self.data[__name]
            return value

        except KeyError:
            raise ValueError(f"Inexistent attribute `{__name}`")

    def __repr__(self) -> str:
        return str(self.data)


class Tickers:
    def __init__(self, tickers: Optional[List[dict]] = []) -> None:
        self.tickers = []

        for ticker in tickers:
            self.add_ticker(ticker)

    def add_ticker(self, ticker: dict) -> None:
        new_ticker = Ticker(ticker)
        self.tickers.append(new_ticker)

    def add_tickers(self, tickers: List[dict]) -> None:
        for ticker in tickers:
            self.add_ticker(ticker)

    def __getitem__(self, idx: int) -> Ticker:
        if idx < len(self.tickers):
            return self.tickers[idx]
        return None

    def __len__(self):
        return len(self.tickers)

    def __repr__(self) -> str:
        return f"Tickers Bin containing {self.__len__()} tickers"

    def to_csv(self, f: str) -> None:
        headers = ','.join([str(i) for i in self.tickers[0]._headers()])
        rows = [headers]

        for ticker in self.tickers:
            rows.append(
                ",".join(
                    [f'"{str(i)}"' if "," in str(i) \
                        or " " in str(i) else str(i) \
                        for i in ticker._values()
                    ]
                ))

        try:
            writer = open(f, "w")

            for row in rows:
                writer.write(str(row))
                writer.write("\n")

            writer.close()

        except Exception as e:
            raise RuntimeError(f"Could not save tickers to csv... {e}")

    def to_list(self) -> List:
        return [ticker._item() for ticker in self.tickers]


from bs4 import BeautifulSoup


def get_number_of_pages(soup: BeautifulSoup) -> int:
    try:
        page_select_button_text = soup.find(
            "select", {"id": "pageSelect"}).find("option").text
        number_of_pages = page_select_button_text.split("/")[-1]

        return number_of_pages
    except:
        return 0


def get_tickers_table(soup: BeautifulSoup) -> dict:

    headers = [item.text for item in soup.find_all(
        "td", {"class": ["table-top", "cursor-pointer"]})]
    table = [item.text for item in soup.find_all(
        "td", {"class": "screener-body-table-nw"})]

    return headers, table
